import re
from typing import ClassVar

import serial
from serial.tools import list_ports

from ... import rosys
from ...geometry import GeoPose, Pose
from .gnss import Gnss, GnssMeasurement
from .nmea import Gga, GpsQuality, Gst, Pssn


class GnssHardware(Gnss):
    """This hardware module connects to a Septentrio SimpleRTK3b (Mosaic-H) GNSS receiver."""
    MAX_MEASUREMENT_AGE = 0.05
    NMEA_TYPES: ClassVar[set[str]] = {'GPGGA', 'GPGST', 'PSSN,HRP'}
    NMEA_PATTERN = re.compile(r'\$(?P<type>[A-Z,]+),(?P<timestamp>\d{6}(?:\.\d+)?)[^*]*\*[0-9A-Fa-f]{2}\r\n')

    def __init__(self, *, antenna_pose: Pose | None, reconnect_interval: float = 3.0) -> None:
        """
        :param antenna_pose: the pose of the main antenna in the robot's coordinate frame (yaw: direction to the auxiliary antenna)
        :param reconnect_interval: the interval to wait before reconnecting to the device
        """
        super().__init__()
        self.antenna_pose = antenna_pose or Pose(x=0.0, y=0.0, yaw=0.0)
        self._reconnect_interval = reconnect_interval
        self.serial_connection: serial.Serial | None = None
        rosys.on_startup(self._run)

    @property
    def is_connected(self) -> bool:
        if self.serial_connection is None:
            self.log.debug('Device not connected')
            return False
        if not self.serial_connection.isOpen():
            self.log.debug('Device not open')
            return False
        return True

    async def _run(self) -> None:
        buffer = ''
        latest_messages: dict[str, tuple[str, str]] = {}  # type -> (timestamp, sentence)
        while True:
            await rosys.sleep(0.01)
            if not self.is_connected and not await self._connect():
                continue
            assert self.serial_connection is not None
            result = self.serial_connection.read_all().decode('utf-8', errors='replace')
            if not result:
                continue
            buffer += result
            matches = list(self.NMEA_PATTERN.finditer(buffer))
            for match in reversed(matches):
                type_, nmea_timestamp = match['type'], match['timestamp']
                if type_ not in self.NMEA_TYPES:
                    self.log.debug('Skipping unknown type: %s', type_)
                    continue
                sentence = match.group(0)
                self.log.debug('%s, %s: %s', type_, nmea_timestamp, sentence)
                sentence = sentence[:sentence.find('*')]
                latest_messages[type_] = (nmea_timestamp, sentence)
                buffer = buffer[:match.start()]
                if not self.NMEA_TYPES.issubset(latest_messages):
                    continue
                timestamps = {latest_messages[msg_type][0] for msg_type in self.NMEA_TYPES}
                if len(timestamps) != 1:
                    latest_timestamp = max(timestamps)
                    latest_messages = {msg_type: (timestamp, sentence)
                                       for msg_type, (timestamp, sentence) in latest_messages.items()
                                       if timestamp >= latest_timestamp}
                    continue
                if not self.NMEA_TYPES.issubset(latest_messages):
                    continue
                try:
                    measurement = self._parse_measurement(latest_messages['GPGGA'][1],
                                                          latest_messages['GPGST'][1],
                                                          latest_messages['PSSN,HRP'][1])
                except ValueError as e:
                    self.log.debug('Failed to parse measurement: %s', e)
                    continue
                measurement_age = measurement.age
                if abs(measurement_age) > self.MAX_MEASUREMENT_AGE:
                    self.log.warning('measurement age = %.3f (exceeds threshold of %s)',
                                     measurement_age, self.MAX_MEASUREMENT_AGE)
                    continue
                self.log.debug('dt: %.3f - %s', measurement_age, measurement)
                self.last_measurement = measurement
                self.NEW_MEASUREMENT.emit(measurement)
                buffer = ''
                latest_messages.clear()
                break

    async def _connect(self) -> bool:
        try:
            serial_device_path = self._find_device()
            self.serial_connection = self._connect_to_device(serial_device_path)
        except RuntimeError:
            self.log.error('Could not connect to GNSS device: %s', serial_device_path)
            await rosys.sleep(self._reconnect_interval)
            return False
        self.log.info('Connected to GNSS device: %s', serial_device_path)
        # NOTE: Allow time for the device to stabilize after connection
        await rosys.sleep(0.1)
        assert self.serial_connection is not None
        self.serial_connection.reset_input_buffer()
        return True

    def _parse_measurement(self, gga_msg: str, gst_msg: str, pssn_msg: str) -> GnssMeasurement:
        gga = Gga.from_sentence(gga_msg)
        gst = Gst.from_sentence(gst_msg)
        pssn = Pssn.from_sentence(pssn_msg)
        last_heading = pssn.heading + self.antenna_pose.yaw_deg
        antenna_pose = GeoPose.from_degrees(gga.latitude, gga.longitude, last_heading)
        robot_pose = antenna_pose.relative_shift_by(x=-self.antenna_pose.x, y=-self.antenna_pose.y)
        return GnssMeasurement(time=rosys.time(),
                               gnss_time=gga.timestamp,
                               pose=robot_pose,
                               latitude_std_dev=gst.latitude_std_dev,
                               longitude_std_dev=gst.longitude_std_dev,
                               heading_std_dev=pssn.heading_std_dev,
                               gps_quality=GpsQuality(gga.gps_quality),
                               num_satellites=gga.num_satellites,
                               hdop=gga.hdop,
                               altitude=gga.altitude)

    def _find_device(self) -> str:
        for port in list_ports.comports():
            self.log.debug('Found port: %s - %s', port.device, port.description)
            if 'Septentrio' in port.description:
                self.log.debug('Found GNSS device: %s', port.device)
                return port.device
        raise RuntimeError('No GNSS device found')

    def _connect_to_device(self, port: str, *, baudrate: int = 921600, timeout: float = 0.2) -> serial.Serial:
        self.log.debug('Connecting to GNSS device "%s"...', port)
        try:
            return serial.Serial(port=port, baudrate=baudrate, timeout=timeout)
        except serial.SerialException as e:
            raise RuntimeError(f'Could not connect to GNSS device: {port}') from e



from collections import defaultdict
from datetime import datetime
from pathlib import Path

from nicegui import run, ui

VIDEO_FILES = Path('~/.rosys/timelapse/videos').expanduser()


class VideosPage:

    def __init__(self) -> None:

        @ui.page('/video/{name:str}')
        def video_page(name: str) -> None:
            self._video_page_content(name)

        @ui.page('/videos', title='Videos')
        async def page():
            await self._content()

    async def _content(self) -> None:
        @ui.refreshable
        def video_ui() -> None:
            with ui.card().props('flat bordered'):
                if date.value not in all_videos:
                    ui.label('No videos for this date')
                    return
                with ui.list().classes('items-center'):
                    for mp4 in all_videos[date.value]:
                        video_date, video_duration = format_video_name(mp4)
                        with ui.item().classes('p-0'):
                            with ui.item_section().props('avatar'):
                                ui.button(icon='download', on_click=lambda mp4=mp4: ui.download(mp4)) \
                                    .props('flat').tooltip('download')
                            with ui.item_section().classes('min-w-24'):
                                with ui.link(target=f'video/{mp4.name}').classes('no-underline'):
                                    ui.item_label(video_date)
                                    ui.item_label(video_duration).props('caption')
                            with ui.item_section().classes('pl-1'):
                                with ui.button(icon='more_vert').props('flat fab-mini'):
                                    with ui.menu():
                                        def delete_video(mp4: Path = mp4) -> None:
                                            mp4.unlink()
                                            all_videos[date.value].remove(mp4)
                                            video_ui.refresh()
                                        with ui.menu_item(on_click=delete_video):
                                            with ui.item_section().props('side'):
                                                ui.icon('delete', color='negative')
                                            ui.item_section('delete')

        ui.label('Timelapse Videos').classes('text-2xl')
        all_videos = await get_all_videos()
        today = datetime.now().strftime(r'%Y%m%d')
        with ui.row(wrap=False).classes('w-full'):
            date = ui.date(value=today, mask='YYYYMMDD', on_change=video_ui.refresh).props('flat bordered')
            date.props['options'] = [datetime.strptime(d, r'%Y%m%d').strftime(r'%Y/%m/%d') for d in all_videos]
            video_ui()

    def _video_page_content(self, name: str) -> None:
        ui.video(VIDEO_FILES / name)


def format_video_name(video: Path) -> tuple[str, str]:
    parts = video.stem.split('_')
    time = parts[1].replace('-', ':')
    return time, f'{parts[2]} {parts[3]}'


async def get_all_videos() -> dict[str, list[Path]]:
    all_videos = sorted(await run.io_bound(VIDEO_FILES.glob, '*.mp4'), reverse=True)
    video_date_dict: dict[str, list[Path]] = defaultdict(list)
    for video in all_videos:
        video_date = video.stem.split('_')[0]
        video_date_dict[video_date].append(video)
    return video_date_dict



from __future__ import annotations

from dataclasses import dataclass
from datetime import UTC, datetime
from enum import IntEnum


class GpsQuality(IntEnum):
    INVALID = 0
    GPS = 1
    DGPS = 2
    PPS = 3
    RTK_FIXED = 4
    RTK_FLOAT = 5
    DEAD_RECKONING = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass(slots=True, kw_only=True)
class Gga:
    """
    GPS fix data and undulation

    https://docs.novatel.com/OEM7/Content/Logs/GPGGA.htm
    """
    timestamp: float
    latitude: float
    longitude: float
    gps_quality: GpsQuality
    num_satellites: int
    hdop: float
    altitude: float
    undulation: float
    correction_age: float

    @staticmethod
    def from_sentence(sentence: str) -> Gga:
        parts = sentence.split(',')
        if not parts[1]:
            raise ValueError(f'Missing timestamp in GGA: {sentence}')
        if parts[2] == '' or parts[4] == '':
            raise ValueError(f'Missing latitude or longitude in GGA: {sentence}')
        return Gga(timestamp=timestamp_from_nmea(parts[1]),
                   latitude=convert_to_decimal(parts[2], parts[3]),
                   longitude=convert_to_decimal(parts[4], parts[5]),
                   gps_quality=GpsQuality(int(parts[6]) if parts[6] else 0),
                   num_satellites=int(parts[7]) if parts[7] else 0,
                   hdop=float(parts[8]) if parts[8] else 0.0,
                   altitude=float(parts[9]) if parts[9] else 0.0,
                   undulation=float(parts[11]) if parts[11] else 0.0,
                   correction_age=float(parts[13]) if parts[13] else 0.0)


@dataclass(slots=True, kw_only=True)
class Gst:
    """
    Estimated error in position solution

    https://docs.novatel.com/OEM7/Content/Logs/GPGST.htm
    """
    timestamp: float
    semimajor_axis_std_dev: float
    semiminor_axis_std_dev: float
    semi_major_axis_heading: float
    latitude_std_dev: float
    longitude_std_dev: float
    altitude_std_dev: float

    @staticmethod
    def from_sentence(sentence: str) -> Gst:
        parts = sentence.split(',')
        if any(part == '' for part in parts[1:9]):
            raise ValueError(f'Missing data in GST: {sentence}')
        return Gst(timestamp=timestamp_from_nmea(parts[1]),
                   semimajor_axis_std_dev=float(parts[3]),
                   semiminor_axis_std_dev=float(parts[4]),
                   semi_major_axis_heading=float(parts[5]),
                   latitude_std_dev=float(parts[6]),
                   longitude_std_dev=float(parts[7]),
                   altitude_std_dev=float(parts[8]))


@dataclass(slots=True, kw_only=True)
class Pssn:
    """PSSN,HRP Septentrio Proprietary Sentence - Heading, Roll, Pitch"""
    timestamp: float
    heading: float
    roll: float
    pitch: float
    heading_std_dev: float
    roll_std_dev: float
    pitch_std_dev: float
    num_satellites: int
    mode: int

    @staticmethod
    def from_sentence(sentence: str) -> Pssn:
        parts = sentence.split(',')
        return Pssn(timestamp=timestamp_from_nmea(parts[2]),
                    heading=float(parts[4] or 0.0),
                    roll=float(parts[5] or 0.0),
                    pitch=float(parts[6] or 0.0),
                    heading_std_dev=float(parts[7] or 'inf'),
                    roll_std_dev=float(parts[8] or 'inf'),
                    pitch_std_dev=float(parts[9] or 'inf'),
                    num_satellites=int(parts[10] or 0),
                    mode=int(parts[11] or 0))


def timestamp_from_nmea(nmea_timestamp: str) -> float:
    today = datetime.now().date()
    time_obj = datetime.strptime(nmea_timestamp, '%H%M%S.%f').time()
    utc_time = datetime.combine(today, time_obj).replace(tzinfo=UTC)
    return utc_time.timestamp()


def convert_to_decimal(coord: str, direction: str) -> float:
    """Convert a coordinate in the format DDMM.mmmm to decimal degrees.

    :param coord: the coordinate to convert
    :param direction: the direction (N/S/E/W)
    :return: the coordinate in decimal degrees
    """
    split_index = coord.find('.') - 2
    degrees = float(coord[:split_index])
    minutes = float(coord[split_index:])
    decimal = degrees + minutes / 60
    if direction in ['S', 'W']:
        decimal = -decimal
    return decimal



import abc

import numpy as np
from nicegui import Event, ui

from .. import helpers, rosys
from ..helpers import remove_indentation
from .bms_message import BmsMessage
from .bms_state import BmsState
from .expander import ExpanderHardware
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class Bms(Module, abc.ABC):
    """The BMS module communicates with a simple battery management system over a serial connection.

    The BMS module provides measured voltages as an event.
    """

    def __init__(self, *, battery_low_threshold: float | None = None, **kwargs) -> None:
        """
        :param battery_low_threshold: If provided, the BATTERY_LOW event will be emitted when the battery percentage gets below this threshold.
        """
        super().__init__(**kwargs)
        self.CHARGING_STARTED = Event[[]]()
        """The battery started charging"""
        self.CHARGING_STOPPED = Event[[]]()
        """The battery stopped charging"""
        self.BATTERY_LOW = Event[[]]()
        """The battery percentage is below the threshold"""

        self.state = BmsState()
        self.raw_data: dict = {}
        self.battery_low_threshold = battery_low_threshold

    def is_above_percent(self, value: float) -> bool:
        """Returns whether the battery is charged above the given percentage."""
        return self.state.percentage is not None and self.state.percentage > value

    def is_below_percent(self, value: float) -> bool:
        """Returns whether the battery is charged below the given percentage."""
        return self.state.percentage is not None and self.state.percentage < value

    def is_above_voltage(self, value: float) -> bool:
        """Returns whether the battery voltage is above the given value."""
        return self.state.voltage is not None and self.state.voltage > value

    def is_below_voltage(self, value: float) -> bool:
        """Returns whether the battery voltage is below the given value."""
        return self.state.voltage is not None and self.state.voltage < value

    def _send_charging_events(self, is_charging: bool) -> None:
        if is_charging and not self.state.is_charging:
            self.CHARGING_STARTED.emit()
        elif not is_charging and self.state.is_charging:
            self.CHARGING_STOPPED.emit()

    def _send_battery_low_event(self, percentage: float) -> None:
        if self.battery_low_threshold is None or self.state.percentage is None:
            return
        if percentage < self.battery_low_threshold and self.state.percentage >= self.battery_low_threshold:
            self.BATTERY_LOW.emit()

    def developer_ui(self) -> None:
        ui.label('Battery Management System').classes('text-center text-bold')
        with ui.grid(columns=2).classes('gap-y-1 [&>*:nth-child(even)]:justify-self-end'):
            ui.label('Percentage:')
            ui.label().bind_text_from(self.state, 'percentage', lambda x: f'{x:.1f}%' if x is not None else 'N/A')
            ui.label('Voltage:')
            ui.label().bind_text_from(self.state, 'voltage', lambda x: f'{x:.1f}V' if x is not None else 'N/A')
            ui.label('Current:')
            ui.label().bind_text_from(self.state, 'current', lambda x: f'{x:.1f}A' if x is not None else 'N/A')
            ui.label('Temperature:')
            ui.label().bind_text_from(self.state, 'temperature', lambda x: f'{x:.1f}°C' if x is not None else 'N/A')
            ui.label('Charging:')
            ui.label().bind_text_from(self.state, 'is_charging', lambda x: 'Yes' if x else 'No')


class BmsHardware(Bms, ModuleHardware):
    """This module implements the hardware interface for the BMS module."""

    UPDATE_INTERVAL = 5.0

    def __init__(self, robot_brain: RobotBrain, *,
                 expander: ExpanderHardware | None = None,
                 name: str = 'bms',
                 rx_pin: int = 26,
                 tx_pin: int = 27,
                 baud: int = 9600,
                 num: int = 1,
                 charge_detect_threshold: float = -0.4,
                 **kwargs,
                 ) -> None:
        """
        :param expander: If provided, the BMS module will use the expander module to create the serial connection.
        :param name: The name of the BMS module.
        :param rx_pin: The GPIO pin number to use for the RX line.
        :param tx_pin: The GPIO pin number to use for the TX line.
        :param baud: The baud rate to use for the serial communication.
        :param num: The serial port index to use for creating the serial connection.
        :param charge_detect_threshold: The threshold current to use for charging detection.
        """
        self.name = name
        self.expander = expander
        self.charge_detect_threshold = charge_detect_threshold
        lizard_code = remove_indentation(f'''
            {name} = {expander.name + "." if expander else ""}Serial({rx_pin}, {tx_pin}, {baud}, {num})
            {name}.unmute()
        ''')
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code, **kwargs)
        rosys.on_repeat(self._request, 1.0)
        self.message_hooks[name] = self._handle_bms

    async def _request(self) -> None:
        if self.robot_brain.is_ready and rosys.time() > self.state.last_update + self.UPDATE_INTERVAL:
            await self.robot_brain.send(f'{self.name}.send(0xdd, 0xa5, 0x03, 0x00, 0xff, 0xfd, 0x77)')

    def _handle_bms(self, line: str) -> None:
        skip = 2 if self.expander else 1
        words = line.split()[skip:]
        msg = BmsMessage([int(w, 16) for w in words])
        msg.check()
        result = msg.interpret()
        percentage = result.get('capacity percent')
        if percentage is not None:
            self._send_battery_low_event(percentage)
        self.state.percentage = percentage
        self.state.voltage = result.get('total voltage')
        self.state.current = result.get('current')
        self.state.temperature = np.mean(result['temperatures']) if 'temperatures' in result else None
        is_charging = (self.state.current or 0) > self.charge_detect_threshold
        self._send_charging_events(is_charging)
        self.state.is_charging = is_charging
        self.state.last_update = rosys.time()
        self.raw_data = result


class BmsSimulation(Bms, ModuleSimulation):
    """This module simulates a BMS module."""

    AVERAGE_VOLTAGE = 25.0
    MIN_VOLTAGE = 22.5
    MAX_VOLTAGE = 27.5
    CHARGING_CURRENT = 1.0
    DISCHARGING_CURRENT = -0.7
    AVERAGE_TEMPERATURE = 20.0
    TEMPERATURE_AMPLITUDE = 1.0
    TEMPERATURE_FREQUENCY = 0.01

    def __init__(self, *, voltage_per_second: float = 0.0, **kwargs) -> None:
        """
        :param voltage_per_second: The voltage change per second. Positive for charging, negative for discharging.
        """
        super().__init__(**kwargs)
        self.voltage_per_second = voltage_per_second
        self.state.voltage = self.AVERAGE_VOLTAGE

    async def step(self, dt: float) -> None:
        assert self.state.voltage is not None
        next_voltage = self.state.voltage + self.voltage_per_second * dt
        self.state.voltage = max(min(next_voltage, self.MAX_VOLTAGE), self.MIN_VOLTAGE)
        is_charging = self.voltage_per_second > 0
        self._send_charging_events(is_charging)
        self.state.is_charging = is_charging
        percentage = helpers.ramp(self.state.voltage, self.MIN_VOLTAGE, self.MAX_VOLTAGE, 0.0, 100.0)
        self._send_battery_low_event(percentage)
        self.state.percentage = percentage
        self.state.current = self.CHARGING_CURRENT if self.state.is_charging else self.DISCHARGING_CURRENT
        self.state.temperature = self.AVERAGE_TEMPERATURE + \
            self.TEMPERATURE_AMPLITUDE * np.sin(self.TEMPERATURE_FREQUENCY * rosys.time())
        self.state.last_update = rosys.time()

    def developer_ui(self) -> None:
        super().developer_ui()
        ui.number(suffix='V/s', step=0.001, format='%.3f').props('dense').classes('w-20') \
            .bind_value(self, 'voltage_per_second') \
            .tooltip('Voltage change per second (positive for charging, negative for discharging)')



from __future__ import annotations

import inspect
import logging
from collections.abc import Callable
from dataclasses import dataclass

from nicegui import Event

from .. import rosys
from ..hardware import BluetoothHardware, EspNotReadyException, RobotBrain
from .automator import Automator


@dataclass
class AppButton:
    icon: str
    state: str = 'enabled'
    visible: bool = True
    pressed: Callable[[], None] = lambda: None
    released: Callable[[], None] = lambda: None
    canceled: Callable[[], None] = lambda: None

    def get_properties(self) -> list[str]:
        return [
            f'/icon {self.icon}',
            f'/state {self.state}',
            f'/visible {"true" if self.visible else "false"}',
        ]

    def __str__(self) -> str:
        return f'AppButton({self.icon},{self.state},{self.visible})'


class AppControls:
    """The AppControls module enables the connection with a mobile-app-based user interface.

    It uses a given RobotBrain object to communicate with [Lizard](https://lizard.dev/) running on a microcontroller
    and in turn being connected to a mobile app via Bluetooth Low Energy.
    It displays buttons to control a given automator.
    """

    def __init__(self, robot_brain: RobotBrain, automator: Automator, bluetooth: BluetoothHardware) -> None:
        self.APP_CONNECTED = Event[[]]()
        """an app connected via bluetooth (used to refresh information or similar)"""

        self.log = logging.getLogger('rosys.app_controls')

        self.automator = automator
        self.bluetooth = bluetooth

        self.main_buttons: dict[str, AppButton] = {
            'play': AppButton('play_arrow', released=automator.start),
            'pause': AppButton('pause', released=lambda: automator.pause(because='pause button was pressed')),
            'resume': AppButton('play_arrow', released=automator.resume),
            'stop': AppButton('stop', released=lambda: automator.stop(because='stop button was pressed')),
        }
        self.extra_buttons: dict[str, AppButton] = {}

        def handle_esp_ready():
            rosys.on_shutdown(self.clear)
            rosys.on_repeat(self.refresh, 0.1)
            rosys.NEW_NOTIFICATION.subscribe(self.notify)
            robot_brain.ESP_CONNECTED.subscribe(self.sync)
            robot_brain.LINE_RECEIVED.subscribe(self.parse)
            robot_brain.ESP_CONNECTED.unsubscribe(handle_esp_ready)

        robot_brain.ESP_CONNECTED.subscribe(handle_esp_ready)

    async def refresh(self) -> None:
        if not self.main_buttons:  # happens on shutdown
            return
        before = ' '.join(str(b) for b in self.main_buttons.values())
        self.main_buttons['play'].visible = self.automator.is_stopped
        self.main_buttons['pause'].visible = self.automator.is_running
        self.main_buttons['resume'].visible = self.automator.is_paused
        self.main_buttons['play'].state = 'disabled' if self.automator.default_automation is None or not self.automator.enabled else 'enabled'
        self.main_buttons['stop'].state = 'disabled' if self.automator.is_stopped else 'enabled'
        after = ' '.join(str(b) for b in self.main_buttons.values())
        if after != before:
            await self.sync()

    async def set_info(self, msg: str) -> None:
        """replace constantly shown info text on mobile device"""
        try:
            await self.bluetooth.send(f'PUT /info {msg}')
        except EspNotReadyException:
            self.log.error('Failed to set info: ESP not ready')

    async def notify(self, msg: str) -> None:
        """show notification as Snackbar message on mobile device"""
        try:
            await self.bluetooth.send(f'POST /notification {msg}')
        except EspNotReadyException:
            self.log.error('Failed to notify: ESP not ready')

    def parse(self, line: str) -> None:
        if line.startswith('"'):
            line = line[1:-1]
        if line.startswith('app: '):
            line = line[5:]
            if line == 'connected':
                rosys.background_tasks.create(self.sync(), name='sync app')
                self.APP_CONNECTED.emit()
            elif line.startswith('PUT /button/') and '/action' in line:
                # line: "PUT /button/main/my_button/action pressed"
                _, path, action = line.split(' ')
                _, group, name, _ = path[1:].split('/')
                buttons = self.main_buttons if group == 'main' else self.extra_buttons
                if action == 'pressed':
                    self._invoke(buttons[name].pressed)
                if action == 'released':
                    self._invoke(buttons[name].released)
                if action == 'canceled':
                    self._invoke(buttons[name].canceled)

    async def sync(self):
        await self.send('PUT', lambda b: b.get_properties())

    async def clear(self):
        await self.send('DELETE')
        self.main_buttons.clear()

    async def send(self, method: str, get_properties: Callable[[AppButton], list[str]] = lambda _: ['']):
        async def run(group: str, buttons: dict[str, AppButton]) -> None:
            for name, button in buttons.items():
                for prop in get_properties(button):
                    await self.bluetooth.send(f'{method} /button/{group}/{name}{prop}')
        try:
            await run('main', self.main_buttons)
            await run('extra', self.extra_buttons)
        except EspNotReadyException:
            self.log.error('Failed to send app controls from %s: ESP not ready', inspect.stack()[1].function)

    def _invoke(self, callback: Callable):
        if inspect.iscoroutinefunction(callback):
            rosys.background_tasks.create(callback(), name='app_controls._invoke')
        else:
            callback()



from datetime import datetime
from pathlib import Path

from fastapi import HTTPException
from fastapi.responses import FileResponse
from nicegui import app, run, ui


class LogsPage:
    """Logs Page

    This module creates a page to list and download log files.
    It is mounted at /logs.
    """

    def __init__(self, *, logs_dir: Path | None = None) -> None:
        self.logs_dir = logs_dir or Path('~/.rosys').expanduser()

        @ui.page('/logs', title='Logs')
        async def page():
            await self._content()

        @app.get('/logs/{name:str}')
        async def download_log(name: str):
            path = self.logs_dir / name
            if not path.exists():
                raise HTTPException(status_code=404, detail=f'Log file {name} not found')
            return FileResponse(path)

    async def _content(self) -> None:
        @ui.refreshable
        def list_ui() -> None:
            with ui.card().tight().props('flat bordered'):
                if logs:
                    with ui.list():
                        for path in logs:
                            with ui.item(on_click=lambda path=path: ui.navigate.to(f'/logs/{path.name}', new_tab=True)):
                                with ui.item_section():
                                    ui.item_label(path.name)
                                    ui.item_label(_file_info(path)).props('caption')
                                with ui.item_section().props('side'):
                                    ui.button(icon='download').on('click.stop', lambda path=path: ui.download(path)) \
                                        .props('flat fab-mini').tooltip('download')
                else:
                    ui.label('No logs found')

        ui.label('Device Logs').classes('text-2xl')
        logs = await _find_log_files(self.logs_dir)
        list_ui()


async def _find_log_files(logs_dir: Path) -> list[Path]:
    logs = await run.io_bound(logs_dir.glob, '*.log')
    rotated = await run.io_bound(logs_dir.glob, '*.log.*')
    paths = list({p.resolve(): p for p in [*logs, *rotated]}.values())
    return sorted(paths, key=lambda p: p.stat().st_mtime, reverse=True)


def _file_info(path: Path) -> str:
    try:
        mtime = datetime.fromtimestamp(path.stat().st_mtime).strftime('%Y-%m-%d %H:%M:%S')
        size = _human_size(path.stat().st_size)
    except FileNotFoundError:
        mtime = 'n/a'
        size = 'n/a'
    return f'{mtime} • {size}'


def _human_size(num_bytes: int) -> str:
    units = ['B', 'KB', 'MB', 'GB', 'TB']
    size = float(num_bytes)
    unit = 0
    while size >= 1024 and unit < len(units) - 1:
        size /= 1024.0
        unit += 1
    return f'{size:.1f} {units[unit]}'



import logging
from typing import Any, Literal, Self

from ..camera.configurable_camera import ConfigurableCamera
from ..camera.transformable_camera import TransformableCamera
from ..image import Image, ImageArray
from ..image_processing import process_ndarray_image
from .rtsp_device import RtspDevice


class RtspCamera(ConfigurableCamera, TransformableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: str | None = None,
                 connect_after_init: bool = True,
                 fps: int = 5,
                 substream: int = 1,
                 bitrate: int = 4096,
                 avdec: Literal['h264', 'h265'] = 'h264',
                 ip: str | None = None,
                 **kwargs) -> None:
        super().__init__(id=id,
                         name=name,
                         connect_after_init=connect_after_init,
                         **kwargs)

        self.log = logging.getLogger(f'rosys.vision.rtsp_camera.{self.id}')

        self.device: RtspDevice | None = None
        self.ip: str | None = ip

        self._register_parameter('substream', self.get_substream, self.set_substream,
                                 min_value=0, max_value=1, step=1, default_value=substream)
        self._register_parameter('fps', self.get_fps, self.set_fps,
                                 min_value=1, max_value=30, step=1, default_value=fps)
        self._register_parameter('bitrate', self.get_bitrate, self.set_bitrate,
                                 min_value=32, max_value=8192, step=1, default_value=bitrate)
        self._register_parameter('avdec', self.get_avdec, self.set_avdec, default_value=avdec)

    def to_dict(self) -> dict[str, Any]:
        parameters = {
            name: param.value for name, param in self._parameters.items()
        }
        return super().to_dict() | parameters | {
            'ip': self.ip,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**cls.args_from_dict(data))

    @property
    def is_connected(self) -> bool:
        return self.device is not None and self.device.is_connected

    @property
    def url(self) -> str | None:
        if not self.is_connected:
            return None
        assert self.device is not None

        return self.device.url

    async def connect(self) -> None:
        if self.is_connected:
            return

        if not self.ip:
            self.log.error('no IP address provided for camera %s', self.id)
            return

        self.device = RtspDevice(mac=self.id, ip=self.ip,
                                 substream=self.parameters['substream'],
                                 fps=self.parameters['fps'],
                                 on_new_image_data=self._handle_new_image_data,
                                 avdec=self.parameters['avdec'])

        await self._apply_all_parameters()

    async def disconnect(self) -> None:
        if not self.is_connected:
            return
        logging.info('camera %s: disconnect initialized...', self.id)

        assert self.device is not None
        await self.device.shutdown()
        self.device = None

    async def _handle_new_image_data(self, image_array: ImageArray, timestamp: float) -> None:
        transformed_image_array = process_ndarray_image(image_array, self.rotation, self.crop)
        image = Image.from_array(transformed_image_array, camera_id=self.id, time=timestamp)
        self._add_image(image)

    async def set_fps(self, fps: int) -> None:
        assert self.device is not None

        await self.device.set_fps(fps)

    async def get_fps(self) -> int | None:
        assert self.device is not None

        return await self.device.get_fps()

    def set_substream(self, index: int) -> None:
        assert self.device is not None

        self.device.set_substream(index)

    def get_substream(self) -> int | None:
        assert self.device is not None

        return self.device.get_substream()

    async def set_bitrate(self, bitrate: int) -> None:
        assert self.device is not None

        await self.device.set_bitrate(bitrate)

    async def get_bitrate(self) -> int | None:
        assert self.device is not None

        return await self.device.get_bitrate()

    def get_avdec(self) -> Literal['h264', 'h265'] | None:
        assert self.device is not None

        return self.device.get_avdec()

    def set_avdec(self, avdec: Literal['h264', 'h265']) -> None:
        assert self.device is not None

        self.device.set_avdec(avdec)

    async def _apply_parameters(self, new_values: dict[str, Any], force_set: bool = False) -> None:
        await super()._apply_parameters(new_values, force_set)
        if self.is_connected:
            assert self.device is not None
            await self.device.restart_gstreamer()



from typing import Literal

from .module import ModuleHardware
from .robot_brain import RobotBrain


class BluetoothHardware(ModuleHardware):

    def __init__(self, robot_brain: RobotBrain, *, name: str = 'robot', pin_code: int | None | Literal['default'] = 'default') -> None:
        """The bluetooth module represents a Bluetooth Low Energy connection for sending messages with optional PIN-based pairing.

        Check https://lizard.dev/module_reference/#bluetooth for more information.

        :param robot_brain: The robot brain instance to communicate with
        :param name: The device name advertised via Bluetooth (default: 'robot')
        :param pin_code: 6-digit PIN (000000-999999) for pairing, ``None`` to deactivate PIN, or "default" for the secret default PIN.
        """
        self.name = name
        lizard_code = f'bluetooth = Bluetooth("{name}")'
        if pin_code != 'default':
            lizard_code += self._create_pin_command(pin_code)
        super().__init__(robot_brain, lizard_code=lizard_code)

    async def send(self, message: str) -> None:
        """Send a message via Bluetooth notification.

        :param message: The message to send. The needed quotation marks will be added automatically.
        """
        await self.robot_brain.send(f'bluetooth.send("{message}")')

    async def set_pin_code(self, pin_code: int | None | Literal['default']) -> None:
        """Set a PIN code for pairing.

        :param pin_code: 6-digit PIN (000000-999999) for pairing, ``None`` to deactivate PIN, or "default" for the secret default PIN.
        """
        await self.robot_brain.send(self._create_pin_command(pin_code))

    async def reset_bonds(self) -> None:
        """Forget all previously paired devices."""
        await self.robot_brain.send('bluetooth.reset_bonds()')

    @staticmethod
    def _create_pin_command(pin_code: int | None | Literal['default']) -> str:
        if pin_code is None:
            return 'bluetooth.deactivate_pin()'
        if pin_code == 'default':
            return 'bluetooth.reset_pin()'
        if 0 <= pin_code <= 999999:
            return f'bluetooth.set_pin({pin_code:06d})'
        raise ValueError(f'Invalid pin code: {pin_code}')



from .geo import Fixpoint, GeoPoint, GeoPose, GeoReference
from .line import Line
from .line_segment import LineSegment
from .object3d import Object3d
from .point import Point
from .point3d import Point3d
from .polygon import Polygon
from .pose import Pose, PoseStep
from .pose3d import AxesObject as axes_object
from .pose3d import Frame3d, Pose3d
from .prism import Prism
from .rectangle import Rectangle
from .rotation import Rotation
from .spline import Spline
from .velocity import Velocity

__all__ = [
    'Fixpoint',
    'Frame3d',
    'GeoPoint',
    'GeoPose',
    'GeoReference',
    'Line',
    'LineSegment',
    'Object3d',
    'Point',
    'Point3d',
    'Polygon',
    'Pose',
    'Pose3d',
    'PoseStep',
    'Prism',
    'Rectangle',
    'Rotation',
    'Spline',
    'Velocity',
    'axes_object',
]
